from apistar.codecs.base import BaseCodec
from apistar.codecs.download import DownloadCodec
from apistar.codecs.jsondata import JSONCodec
from apistar.codecs.jsonschema import JSONSchemaCodec
from apistar.codecs.multipart import MultiPartCodec
from apistar.codecs.openapi import OpenAPICodec
from apistar.codecs.text import TextCodec
from apistar.codecs.urlencoded import URLEncodedCodec

__all__ = [
    'BaseCodec', 'JSONCodec', 'JSONSchemaCodec', 'OpenAPICodec', 'TextCodec',
    'DownloadCodec', 'MultiPartCodec', 'URLEncodedCodec',
]



from apistar.codecs.base import BaseCodec


class TextCodec(BaseCodec):
    media_type = 'text/*'
    format = 'text'

    def decode(self, bytestring, **options):
        return bytestring.decode('utf-8')



import re
from json.decoder import JSONDecodeError, JSONDecoder, scanstring

from apistar.formatters.nodes import DictNode, ListNode, LiteralNode

FLAGS = re.VERBOSE | re.MULTILINE | re.DOTALL
WHITESPACE = re.compile(r'[ \t\n\r]*', FLAGS)
WHITESPACE_STR = ' \t\n\r'
NUMBER_RE = re.compile(
    r'(-?(?:0|[1-9]\d*))(\.\d+)?([eE][-+]?\d+)?',
    (re.VERBOSE | re.MULTILINE | re.DOTALL))


def JSONObject(s_and_end, strict, scan_once, object_hook, object_pairs_hook,
               memo=None, _w=WHITESPACE.match, _ws=WHITESPACE_STR):
    s, end = s_and_end
    pairs = []
    pairs_append = pairs.append
    # Backwards compatibility
    if memo is None:
        memo = {}
    memo_get = memo.setdefault
    # Use a slice to prevent IndexError from being raised, the following
    # check will raise a more specific ValueError if the string is empty
    nextchar = s[end:end + 1]
    # Normally we expect nextchar == '"'
    if nextchar != '"':
        if nextchar in _ws:
            end = _w(s, end).end()
            nextchar = s[end:end + 1]
        # Trivial empty object
        if nextchar == '}':
            if object_pairs_hook is not None:
                result = object_pairs_hook(pairs)
                return result, end + 1
            pairs = {}
            if object_hook is not None:
                pairs = object_hook(pairs)
            return pairs, end + 1
        elif nextchar != '"':
            raise JSONDecodeError(
                "Expecting property name enclosed in double quotes", s, end)
    end += 1
    while True:
        key_start = end - 1
        key, end = scanstring(s, end, strict)
        key_end = end
        key = memo_get(key, key)
        # To skip some function call overhead we optimize the fast paths where
        # the JSON key separator is ": " or just ":".
        if s[end:end + 1] != ':':
            end = _w(s, end).end()
            if s[end:end + 1] != ':':
                raise JSONDecodeError("Expecting ':' delimiter", s, end)
        end += 1

        try:
            if s[end] in _ws:
                end += 1
                if s[end] in _ws:
                    end = _w(s, end + 1).end()
        except IndexError:
            pass

        try:
            value, end = scan_once(s, end)
        except StopIteration as err:
            raise JSONDecodeError("Expecting value", s, err.value) from None
        value.key_start = key_start
        value.key_end = key_end
        pairs_append((key, value))
        try:
            nextchar = s[end]
            if nextchar in _ws:
                end = _w(s, end + 1).end()
                nextchar = s[end]
        except IndexError:
            nextchar = ''
        end += 1

        if nextchar == '}':
            break
        elif nextchar != ',':
            raise JSONDecodeError("Expecting ',' delimiter", s, end - 1)
        end = _w(s, end).end()
        nextchar = s[end:end + 1]
        end += 1
        if nextchar != '"':
            raise JSONDecodeError(
                "Expecting property name enclosed in double quotes", s, end - 1)
    if object_pairs_hook is not None:
        result = object_pairs_hook(pairs)
        return result, end
    pairs = dict(pairs)
    if object_hook is not None:
        pairs = object_hook(pairs)
    return pairs, end


def make_scanner(context):
    parse_object = JSONObject
    parse_array = context.parse_array
    parse_string = context.parse_string
    match_number = NUMBER_RE.match
    strict = context.strict
    parse_float = context.parse_float
    parse_int = context.parse_int
    parse_constant = context.parse_constant
    object_hook = context.object_hook
    object_pairs_hook = context.object_pairs_hook
    memo = context.memo

    def _scan_once(string, idx):
        try:
            nextchar = string[idx]
        except IndexError:
            raise StopIteration(idx) from None

        if nextchar == '"':
            value, end = parse_string(string, idx + 1, strict)
            return LiteralNode(value, idx, end), end
        elif nextchar == '{':
            value, end = parse_object(
                (string, idx + 1), strict,
                _scan_once, object_hook, object_pairs_hook, memo
            )
            return DictNode(value, idx, end), end
        elif nextchar == '[':
            value, end = parse_array((string, idx + 1), _scan_once)
            return ListNode(value, idx, end), end
        elif nextchar == 'n' and string[idx:idx + 4] == 'null':
            value, end = None, idx + 4
            return LiteralNode(value, idx, end), end
        elif nextchar == 't' and string[idx:idx + 4] == 'true':
            value, end = True, idx + 4
            return LiteralNode(value, idx, end), end
        elif nextchar == 'f' and string[idx:idx + 5] == 'false':
            value, end = False, idx + 5
            return LiteralNode(value, idx, end), end

        m = match_number(string, idx)
        if m is not None:
            integer, frac, exp = m.groups()
            if frac or exp:
                res = parse_float(integer + (frac or '') + (exp or ''))
            else:
                res = parse_int(integer)
            value, end = res, m.end()
            return LiteralNode(value, idx, end), end
        elif nextchar == 'N' and string[idx:idx + 3] == 'NaN':
            value, end = parse_constant('NaN'), idx + 3
            return LiteralNode(value, idx, end), end
        elif nextchar == 'I' and string[idx:idx + 8] == 'Infinity':
            value, end = parse_constant('Infinity'), idx + 8
            return LiteralNode(value, idx, end), end
        elif nextchar == '-' and string[idx:idx + 9] == '-Infinity':
            value, end = parse_constant('-Infinity'), idx + 9
            return LiteralNode(value, idx, end), end
        else:
            raise StopIteration(idx)

    def scan_once(string, idx):
        try:
            return _scan_once(string, idx)
        finally:
            memo.clear()

    return _scan_once


class JSONNodeDecoder(JSONDecoder):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.scan_once = make_scanner(self)


def parse_json(content):
    decoder = JSONNodeDecoder()
    return decoder.decode(content)



from typing import Union


class ValidationError(Exception):
    def __init__(self, detail, value=None):
        assert isinstance(detail, (str, dict))
        self.detail = detail
        self.value = value
        super(ValidationError, self).__init__(detail)


class ParseError(Exception):
    """
    Raised by a Codec when `decode` fails due to malformed syntax.
    """
    def __init__(self, message, short_message=None, pos=None, lineno=None, colno=None):
        Exception.__init__(self, message)
        self.message = message
        self.short_message = short_message
        self.pos = pos
        self.lineno = lineno
        self.colno = colno


class NoReverseMatch(Exception):
    """
    Raised by a Router when `reverse_url` is passed an invalid handler name.
    """
    pass


class ErrorResponse(Exception):
    """
    Raised when a client request results in an error response being returned.
    """
    def __init__(self, title, content):
        self.title = title
        self.content = content


class RequestError(Exception):
    """
    Raised when some invalid parameter is used in a client request.
    """
    pass


class NoCodecAvailable(Exception):
    pass


class ConfigurationError(Exception):
    pass


# HTTP exceptions

class HTTPException(Exception):
    default_status_code = None  # type: int
    default_detail = None  # type: str

    def __init__(self,
                 detail: Union[str, dict]=None,
                 status_code: int=None) -> None:
        self.detail = self.default_detail if (detail is None) else detail
        self.status_code = self.default_status_code if (status_code is None) else status_code
        assert self.detail is not None, '"detail" is required.'
        assert self.status_code is not None, '"status_code" is required.'

    def get_headers(self):
        return {}


class Found(HTTPException):
    default_status_code = 302
    default_detail = 'Found'

    def __init__(self,
                 location: str,
                 detail: Union[str, dict]=None,
                 status_code: int=None) -> None:
        self.location = location
        super().__init__(detail, status_code)

    def get_headers(self):
        return {'Location': self.location}


class BadRequest(HTTPException):
    default_status_code = 400
    default_detail = 'Bad request'


class Forbidden(HTTPException):
    default_status_code = 403
    default_detail = 'Forbidden'


class NotFound(HTTPException):
    default_status_code = 404
    default_detail = 'Not found'


class MethodNotAllowed(HTTPException):
    default_status_code = 405
    default_detail = 'Method not allowed'


class NotAcceptable(HTTPException):
    default_status_code = 406
    default_detail = 'Could not satisfy the request Accept header'


class UnsupportedMediaType(HTTPException):
    default_status_code = 415
    default_detail = 'Unsupported Content-Type header in request'



import asyncio
import io
import typing
from urllib.parse import unquote, urlparse

import requests


class _HeaderDict(requests.packages.urllib3._collections.HTTPHeaderDict):
    def get_all(self, key, default):
        return self.getheaders(key)


class _MockOriginalResponse(object):
    """
    We have to jump through some hoops to present the response as if
    it was made using urllib3.
    """
    def __init__(self, headers):
        self.msg = _HeaderDict(headers)
        self.closed = False

    def isclosed(self):
        return self.closed

    def close(self):
        self.closed = True


class _WSGIAdapter(requests.adapters.HTTPAdapter):
    """
    A transport adapter for `requests` that makes requests directly to a
    WSGI app, rather than making actual HTTP requests over the network.
    """
    def __init__(self, app: typing.Callable) -> None:
        self.app = app

    def get_environ(self, request: requests.PreparedRequest) -> typing.Dict[str, typing.Any]:
        """
        Given a `requests.PreparedRequest` instance, return a WSGI environ dict.
        """
        body = request.body
        if isinstance(body, str):
            body_bytes = body.encode("utf-8")  # type: bytes
        else:
            body_bytes = body

        url_components = urlparse(request.url)
        environ = {
            'REQUEST_METHOD': request.method,
            'wsgi.url_scheme': url_components.scheme,
            'SCRIPT_NAME': '',
            'PATH_INFO': unquote(url_components.path),
            'wsgi.input': io.BytesIO(body_bytes),
        }  # type: typing.Dict[str, typing.Any]

        if url_components.query:
            environ['QUERY_STRING'] = url_components.query

        if url_components.port:
            environ['SERVER_NAME'] = url_components.hostname
            environ['SERVER_PORT'] = str(url_components.port)
        else:
            environ['HTTP_HOST'] = url_components.hostname

        for key, value in request.headers.items():
            key = key.upper().replace('-', '_')
            if key not in ('CONTENT_LENGTH', 'CONTENT_TYPE'):
                key = 'HTTP_' + key
            environ[key] = value

        return environ

    def send(self, request, *args, **kwargs):
        """
        Make an outgoing request to a WSGI application.
        """
        raw_kwargs = {}

        def start_response(wsgi_status, wsgi_headers, exc_info=None):
            if exc_info is not None:
                raise exc_info[0].with_traceback(exc_info[1], exc_info[2])
            status, _, reason = wsgi_status.partition(' ')
            raw_kwargs['status'] = int(status)
            raw_kwargs['reason'] = reason
            raw_kwargs['headers'] = wsgi_headers
            raw_kwargs['version'] = 11
            raw_kwargs['preload_content'] = False
            raw_kwargs['original_response'] = _MockOriginalResponse(wsgi_headers)

        # Make the outgoing request via WSGI.
        environ = self.get_environ(request)
        wsgi_response = self.app(environ, start_response)

        # Build the underlying urllib3.HTTPResponse
        raw_kwargs['body'] = io.BytesIO(b''.join(wsgi_response))
        raw = requests.packages.urllib3.HTTPResponse(**raw_kwargs)

        # Build the requests.Response
        return self.build_response(request, raw)


class _ASGIAdapter(requests.adapters.HTTPAdapter):
    def __init__(self, app: typing.Callable) -> None:
        self.app = app

    def send(self, request, *args, **kwargs):
        scheme, netloc, path, params, query, fragement = urlparse(request.url)
        if ':' in netloc:
            host, port = netloc.split(':', 1)
            port = int(port)
        else:
            host = netloc
            port = {'http': 80, 'https': 443}[scheme]

        # Include the 'host' header.
        if 'host' in request.headers:
            headers = []
        elif port == 80:
            headers = [[b'host', host.encode()]]
        else:
            headers = [[b'host', ('%s:%d' % (host, port)).encode()]]

        # Include other request headers.
        headers += [
            [key.encode(), value.encode()]
            for key, value in request.headers.items()
        ]

        scope = {
            'type': 'http',
            'http_version': '1.1',
            'method': request.method,
            'path': unquote(path),
            'root_path': '',
            'scheme': scheme,
            'query_string': query.encode(),
            'headers': headers,
            'client': ['testclient', 50000],
            'server': [host, port],
            'raise_exceptions': True  # Not actually part of the spec.
        }

        async def receive():
            body = request.body
            if isinstance(body, str):
                body_bytes = body.encode("utf-8")  # type: bytes
            elif body is None:
                body_bytes = b''
            else:
                body_bytes = body
            return {
                'type': 'http.request',
                'body': body_bytes,
            }

        async def send(message):
            if message['type'] == 'http.response.start':
                raw_kwargs['version'] = 11
                raw_kwargs['status'] = message['status']
                raw_kwargs['headers'] = [
                    (key.decode(), value.decode())
                    for key, value in message['headers']
                ]
                raw_kwargs['preload_content'] = False
                raw_kwargs['original_response'] = _MockOriginalResponse(raw_kwargs['headers'])
            elif message['type'] == 'http.response.body':
                raw_kwargs['body'] = io.BytesIO(message['body'])
            elif message['type'] == 'http.disconnect':
                pass
            elif message['type'] == 'http.exc_info':
                exc_info = message['exc_info']
                raise exc_info[0].with_traceback(exc_info[1], exc_info[2])
            else:
                raise Exception("Unknown ASGI message type: %s" % message['type'])

        raw_kwargs = {}
        connection = self.app(scope)

        loop = asyncio.get_event_loop()
        loop.run_until_complete(connection(receive, send))

        raw = requests.packages.urllib3.HTTPResponse(**raw_kwargs)
        return self.build_response(request, raw)


class _TestClient(requests.Session):
    def __init__(self, app: typing.Callable, scheme: str, hostname: str) -> None:
        super(_TestClient, self).__init__()
        interface = getattr(app, 'interface', None)
        if interface == 'asgi':
            adapter = _ASGIAdapter(app)
        else:
            adapter = _WSGIAdapter(app)
        self.mount('http://', adapter)
        self.mount('https://', adapter)
        self.headers.update({'User-Agent': 'testclient'})
        self.scheme = scheme
        self.hostname = hostname

    def request(self, method: str, url: str, **kwargs) -> requests.Response:  # type: ignore
        if not (url.startswith('http:') or url.startswith('https:')):
            assert url.startswith('/'), (
                "TestClient expected either "
                "an absolute URL starting 'http:' / 'https:', "
                "or a relative URL starting with '/'. URL was '%s'." % url
            )
            url = '%s://%s%s' % (self.scheme, self.hostname, url)
        return super().request(method, url, **kwargs)


def TestClient(app: typing.Callable, scheme: str='http', hostname: str='testserver') -> _TestClient:
    """
    We have to work around py.test discovery attempting to pick up
    the `TestClient` class, by declaring this as a function.
    """
    return _TestClient(app, scheme, hostname)
